from __future__ import annotations
from typing import Callable


class EstimatorBuilder(object):
    """
    Examples
    --------

    estimator = EstimatorBuilder.of(class).parameters(dict).fit_with(fit_func).predict_with(predict_func).build()
    
    """

    def of(self, model_type) -> EstimatorBuilder:
        """
        Assign a model type for the estimator

        Parameters:
            model_type (?): A model type

        Returns:
            (EstimatorBuilder): The current object
        """

        self._model_type = model_type
        return self

    def fit_with(self, func: Callable) -> EstimatorBuilder:
        """
        Define a function that will be used for the model training

        Parameters:
            func (Callable): A fit function used for model training

        Returns:
            (EstimatorBuilder): The current object
        """

        self._fit = func
        return self

    def predict_with(self, func: Callable) -> EstimatorBuilder:
        """
        Define a function that will be used for the model inference

        Parameters:
            func (Callable): A predict function used for model inference

        Returns:
            (EstimatorBuilder): The current object
        """

        self._predict = func
        return self

    def build(self) -> BaseEstimator:
        """
        Creates an instance of BaseEstimator

        Returns:
            (BaseEstimator): An instance of BaseEstimator that will be used for the optimization
        """

        return BaseEstimator(model_type=self._model_type, fit_func=self._fit, predict_func=self._predict)


class BaseEstimator(object):

    """
    A base class to be used for model optimization
    """

    def __init__(self, model_type, fit_func: Callable, predict_func: Callable) -> None:
        """
        Create a class instance

        Parameters:
            model_type (?): A model type
            parameters (dict): Possible model parameters
            fit_func (Callable): A fit function used for model training
            predict_func (Callable): A predict function used for model inference
        """
        
        self._parameters = None
        self._model = None
        self._model_type = model_type
        self._fit_func = fit_func
        self._predict_func = predict_func
        self._fitness = -1

    def initialize(self, parameters: dict) -> BaseEstimator:
        """
        Create a class instance

        Parameters:
            parameters (dict): Possible model parameters
        Returns:
            (BaseEstimator): The current object with the model initialized
        """

        self._parameters = parameters
        self._model = self._model_type(**self._parameters)
        return self

    @property
    def model_type(self):
        self._model_type

    @property
    def parameters(self) -> dict:
        """
        Property to access the base model parameters

        Returns:
            (dict): The parameters
        """
        return self._parameters

    @property
    def model(self):
        """
        Property to create a model from the model type and parameters

        Returns:
            (?): The base model
        """

        return self._model

    @property
    def fitness(self) -> float:
        """
        Property to access the base model metric value

        Returns:
            (float): The metric value
        """

        return self._fitness

    @fitness.setter
    def fitness(self, value) -> None:
        """
        Setter

        Parameters:
            value (float): The new fitness value
        """
        self._fitness = value

    def fit(self, x, y) -> None:
        """
        Method that performs model training. The inheriting class must implement this method.

        Parameters:
            x (?): The data set
            y (?): The correct label set
        """
        self._model = self._fit_func(self._model, x, y)

    def predict(self, x) -> list:
        """
        Method that performs model inference. The inheriting class must implement this method.

        Parameters:
            x (?): The data set

        Returns:
            (list): The predict output
        """
        return self._predict_func(self._model, x)



"""
Class that holds a genetic algorithm for evolving a network.

Credit:
    A lot of those code was originally inspired by:
    https://github.com/mrpeel/genetic-keras
"""

from typing import List, Callable, Type, TypeVar
from abc import ABC, abstractmethod
from tqdm import tqdm
from geneticml.strategy import BaseStrategy
from geneticml.algorithms import BaseEstimator

T = TypeVar('T', bound=BaseEstimator)
E = TypeVar('E', bound=BaseStrategy)


class BaseOptimizer(ABC):
    def __init__(self, strategy: Type[BaseStrategy]) -> None:
        super().__init__()
        pass

    @abstractmethod
    def simulate(self, data, target, verbose: bool = True) -> List[T]:
        """
        Generate a network with the genetic algorithm.

        Parameters:
            data (?): The data used to train the algorithm
            target (?): The targets used to train the algorithm
            verbose (bool): True if should verbose or False if not

        Returns:
            (List[BaseEstimator]): A list with the final population sorted by their loss
        """
        pass


class GeneticOptimizer(BaseOptimizer):
    def __init__(self, strategy: Type[BaseStrategy]) -> None:
        """
        Create an optimizer.

        Parameters:
            strategy (Type[BaseStrategy]): Any strategy that inherits from the type defined. This strategy will be used for the optimization
        """
        super().__init__(strategy)
        self._strategy = strategy

    def simulate(self, data, target, generations: int, population: int, evaluation_function: Callable, greater_is_better: bool = False, verbose: bool = True) -> List[T]:
        """
        Generate a network with the genetic algorithm.

        Parameters:
            generations (int): Number of times to evole the population
            population (int): Number of estimators in each generation
            evaluation_function (Callable): The function that will calculate the metric
            greater_is_better (bool) Whether evaluation_function is a score function (default), meaning high is good, or a loss function, meaning low is good. In the latter case, the scorer object will sign-flip the outcome of the evaluation_function.
            verbose (bool): True if should verbose or False if not

        Returns:
            (List[BaseEstimator]): A list with the final population sorted by their loss
        """

        estimators = self._strategy.create_population(population)

        if verbose:
            increment = 100 / generations
            pbar = tqdm(total=100)

        # Evolve the generation.
        for i in range(generations):

            # Train and get loss for networks.
            losses = []
            for x in estimators:

                # Train the model
                x.fit(data, target)

                # Do the model inference
                y_pred = x.predict(data)

                # Peform the evaluation
                loss = evaluation_function(target, y_pred)

                # Workaround to make the lower losses better than the bigger ones
                if greater_is_better:
                    loss = 1 - loss

                # Assign the evaluation result to the estimator
                x.fitness = loss

                # Track the loss
                losses.append(loss)

            avgloss = sum(losses) / len(losses)
            avgloss = avgloss if greater_is_better else 1 - avgloss

            # Evolve, except on the last iteration.
            if i != generations - 1:
                # Do the evolution.
                estimators = self._strategy.execute(estimators)
        
            if verbose:
                pbar.update(increment)
                pbar.set_postfix({'generation_loss': avgloss})
        
        if verbose:
            pbar.close()

        # Sort our final population.
        return sorted(estimators, key=lambda x: x.fitness, reverse=False)



"""
Class that holds a genetic algorithm for evolving a estimator.

Credit:
    A lot of those code was originally inspired by:
    https://github.com/mrpeel/genetic-keras
"""

from abc import ABC, abstractmethod
from typing import List, Type, TypeVar
from functools import reduce
from operator import add
import random
from geneticml.algorithms import BaseEstimator

T = TypeVar('T', bound=BaseEstimator)


class BaseStrategy(ABC):
    def __init__(self, estimator_type: Type[BaseEstimator]) -> None:
        super().__init__()
        pass

    @abstractmethod
    def execute(self, population: List[Type[T]]) -> List[T]:
        """
        Execute the strategy on a population

        Parameters:
            population (list): A list of estimator parameters

        Returns:
            (list): The population
        """
        pass


class EvolutionaryStrategy(BaseStrategy):
    """Class that implements genetic algorithm for MLP optimization."""

    def __init__(self, estimator_type: Type[BaseEstimator], parameters: dict, retain: float = 0.4, random_select: float = 0.1, mutate_chance: float = 0.2, max_children: int = 2, random_state: int = 1231) -> None:
        """
        Create an optimizer.

        Parameters:
            estimator_type (Type[BaseEstimator]): Any instance that inherits from the type defined
            parameters (dict): Possible model paremters
            retain (float): Percentage of population to retain after each generation
            random_select (float): Probability of a rejected estimator remaining in the population
            mutate_chance (float): Probability a estimator will be randomly mutated
            max_children (int): The maximum size of babies that every family could have
            random_state (int): The random state used as seed for the algorithms
        """
        super().__init__(estimator_type)
        self._estimator_type = estimator_type
        self.mutate_chance = mutate_chance
        self.random_select = random_select
        self.retain = retain
        self._parameters = parameters
        self._max_children = max_children
        self._random_state = random_state

        random.seed(self._random_state)

    def create_random_set(self) -> dict:
        """
        Generate a random set of model parameters

        Returns:
            (dict): A new set of parameters
        """

        # Define the seed
        random.seed(self._random_state)

        params = {}
        for key in self._parameters:
            params[key] = random.choice(self._parameters[key])
        return params

    def create_population(self, size: int) -> List[T]:
        """
        Create a population of random networks.

        Parameters:
            size (int): Number of networks to generate, aka the size of the population

        Returns:
            (list): Population of algorithms.BaseAlgorithm objects
        """

        return [self._estimator_type.initialize(self.create_random_set()) for val in range(0, size)]

    def grade(self, population: list) -> float:
        """
        Find average fitness for a population.

        Parameters:
            population (list): The population of networks

        Returns:
            (float): The average loss of the population
        """
        summed = reduce(add, (estimator.fitness for estimator in population))
        return summed / float((len(population)))

    def breed(self, parent1: T, parent2: T) -> List[T]:
        """
        Make children as parts of their parents.

        Parameters:
            parent1 (dict): The model parameters
            parent2 (dict): The model parameters

        Returns:
            (List[BaseEstimator]): Estimator objects
        """

        # Define the seed
        random.seed(self._random_state)

        children = random.randint(1, self._max_children)

        return [self._estimator_type.initialize(parameters={param: random.choice([parent1.parameters[param], parent2.parameters[param]]) for param in self._parameters}) for _ in range(0, children)]

    def mutate(self, parameters: dict) -> dict:
        """
        Randomly mutate one part of the parameters.

        Parameters:
            parameters (dict): The model parameters to mutate

        Returns:
            (dict): A randomly mutated parameters

        """

        # Define the seed
        random.seed(self._random_state)

        # Choose a random key.
        mutation = random.choice(list(self._parameters.keys()))

        # Mutate one of the params.
        parameters[mutation] = random.choice(self._parameters[mutation])

        return parameters

    def execute(self, population: List[Type[T]]) -> List[T]:
        """
        Evolve a population of networks.

        Parameters:
            population (list): A list of estimator parameters

        Returns:
            (list): The evolved population of networks
        """

        # Define the seed
        random.seed(self._random_state)

        # Get scores for each estimator.
        graded = [(estimator.fitness, estimator) for estimator in population]

        # Sort on the scores.
        graded = [x[1] for x in sorted(graded, key=lambda x: x[0], reverse=False)]

        # Get the number we want to keep for the next gen.
        retain_length = int(len(graded) * self.retain)

        # The parents are every estimator we want to keep.
        parents = graded[:retain_length]

        # For those we aren't keeping, randomly keep some anyway.
        for individual in graded[retain_length:]:
            if self.random_select > random.random():
                parents.append(individual)

        # Randomly mutate some of the networks we're keeping.
        for i in range(len(parents)):
            if self.mutate_chance > random.random():
                parents[i] = self._estimator_type.initialize(parameters=self.mutate(parents[i].parameters))

        # Now find out how many spots we have left to fill.
        parents_length = len(parents)
        desired_length = len(population) - parents_length
        children = []

        # Add children, which are bred from two remaining networks.
        while len(children) < desired_length:

            # Get a random mom and dad.
            parent1 = random.randint(0, parents_length - 1)
            parent2 = random.randint(0, parents_length - 1)

            # Assuming they aren't the same estimator...
            if parent1 != parent2:
                parent1 = parents[parent1]
                parent2 = parents[parent2]

                # Breed them.
                babies = self.breed(parent1, parent2)

                # Add the children one at a time.
                for baby in babies:
                    # Don't grow larger than desired length.
                    if len(children) < desired_length:
                        children.append(baby)

        parents.extend(children)

        return parents



from setuptools import setup, find_packages

def readme():
    with open('README.md') as f:
        return f.read()
 
setup(
    name='geneticml',
    version='1.0.3',
    description='A simple and lightweight genetic algorithm for optimization of any machine learning model',
    long_description=readme(),
    long_description_content_type="text/markdown",
    url='https://github.com/albarsil/geneticml',
    author='Allan Barcelos',
    author_email='albarsil@gmail.com',
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ],
    keywords=['machine learning', 'genetic', 'sklearn', 'pytorch', 'data science'],
    packages=find_packages(exclude=['tests.*', 'tests']),
    install_requires=['tqdm'],
    test_suite='tests.test_suite'
)
